"""byte-pysqlite - executor connection module."""
from __future__ import absolute_import, division, print_function

from byte.executors.core.models.database import DatabaseConnection


class PySqliteConnection(DatabaseConnection):
    """SQLite connection class."""

    def __init__(self, executor, instance):
        """Create sqlite connection.

        :param executor: Executor
        :type executor: byte.executors.core.base.Executor

        :param instance: SQLite Connection
        :type instance: sqlite3.Connection
        """
        super(PySqliteConnection, self).__init__(executor)

        self.instance = instance



"""byte-pysqlite - executor module."""
from __future__ import absolute_import, division, print_function

from byte.executors.core.base import DatabaseExecutorPlugin
from byte.executors.pysqlite.models.connection import PySqliteConnection
from byte.executors.pysqlite.models.cursor import PySqliteCursor
from byte.executors.pysqlite.models.transaction import PySqliteTransaction
from byte.executors.pysqlite.tasks import PySqliteInsertTask, PySqliteSelectTask
from byte.queries import InsertQuery, SelectQuery

from pysqlite2 import dbapi2 as sqlite3
import logging
import os

log = logging.getLogger(__name__)


class PySqliteExecutor(DatabaseExecutorPlugin):
    """PySQLite executor class."""

    key = 'pysqlite'

    class Meta(DatabaseExecutorPlugin.Meta):
        """PySQLite executor metadata."""

        content_type = 'application/x-sqlite3'

        extension = [
            'db',
            'sqlite'
        ]

        scheme = [
            'pysqlite',
            'sqlite'
        ]

    @property
    def path(self):
        """Retrieve database path.

        :return: Database Path
        :rtype: str
        """
        path = (
            self.collection.uri.netloc +
            self.collection.uri.path
        )

        if path == ':memory:':
            return path

        return os.path.abspath(path)

    def construct_compiler(self):
        """Construct compiler."""
        return self.plugins.get_compiler('sqlite')(
            self,
            version=sqlite3.sqlite_version_info
        )

    def create_connection(self):
        """Connect to database.

        :return: SQLite Connection
        :rtype: sqlite3.Connection
        """
        # Connect to database
        instance = sqlite3.connect(self.path)
        instance.isolation_level = None

        # Create connection
        connection = PySqliteConnection(self, instance)

        # Configure connection
        with connection.transaction() as transaction:
            # Enable write-ahead logging
            transaction.execute('PRAGMA journal_mode=WAL;')

        return connection

    def create_transaction(self, connection=None):
        """Create database transaction.

        :return: SQLite Connection
        :rtype: sqlite3.Connection
        """
        return PySqliteTransaction(
            self,
            connection=connection
        )

    def cursor(self, connection=None):
        """Create database cursor.

        :return: Cursor
        :rtype: byte.executors.sqlite.models.cursor.SqliteCursor
        """
        return PySqliteCursor(
            self,
            connection=connection
        )

    def execute(self, query):
        """Execute query.

        :param query: Query
        :type query: byte.queries.Query
        """
        statements = self.compiler.compile(query)

        if not statements:
            raise ValueError('No statements returned from compiler')

        # Construct task
        if isinstance(query, SelectQuery):
            return PySqliteSelectTask(self, statements).execute()

        if isinstance(query, InsertQuery):
            return PySqliteInsertTask(self, statements).execute()

        raise NotImplementedError('Unsupported query: %s' % (type(query).__name__,))



from __future__ import absolute_import, division, print_function

from byte.collection import Collection
from byte.model import Model
from byte.property import Property
import byte.compilers.sqlite
import byte.executors.pysqlite

from hamcrest import *


class User(Model):
    class Options:
        slots = True

    id = Property(int, primary_key=True)

    username = Property(str)
    password = Property(str)


def test_all():
    """Test all items can be retrieved from database."""
    users = Collection(User, 'pysqlite://:memory:?table=users', plugins=[
        byte.compilers.sqlite,
        byte.executors.pysqlite
    ])

    # Create table, and add items directly to database
    with users.executor.connection() as connection:
        with connection.cursor() as cursor:
            cursor.execute("""
                CREATE TABLE users (
                    id          INTEGER         PRIMARY KEY AUTOINCREMENT NOT NULL,
                    username    VARCHAR(255),
                    password    VARCHAR(255)
                );
            """)

            cursor.execute("INSERT INTO users (username, password) VALUES ('one', 'alpha');")
            cursor.execute("INSERT INTO users (username, password) VALUES ('two', 'beta');")
            cursor.execute("INSERT INTO users (username, password) VALUES ('three', 'charlie');")

    # Validate items
    assert_that(users.all(), only_contains(
        has_properties({
            'username': 'one',
            'password': 'alpha'
        }),
        has_properties({
            'username': 'two',
            'password': 'beta'
        }),
        has_properties({
            'username': 'three',
            'password': 'charlie'
        })
    ))



"""byte-pysqlite - executor transaction module."""
from __future__ import absolute_import, division, print_function

from byte.executors.core.models.database import DatabaseTransaction
from byte.executors.pysqlite.models.cursor import PySqliteCursor


class PySqliteTransaction(DatabaseTransaction, PySqliteCursor):
    """PySQLite transaction class."""

    def begin(self):
        """Begin transaction."""
        self.instance.execute('BEGIN;')

    def commit(self):
        """Commit transaction."""
        self.connection.instance.commit()

    def rollback(self):
        """Rollback transaction."""
        self.connection.instance.rollback()

    def close(self):
        """Close transaction."""
        # Close cursor
        self.instance.close()
        self.instance = None

        # Close transaction
        super(PySqliteTransaction, self).close()



"""byte-pysqlite - executor cursor module."""
from __future__ import absolute_import, division, print_function

from byte.executors.core.models.database import DatabaseCursor

import logging

log = logging.getLogger(__name__)


class PySqliteCursor(DatabaseCursor):
    """SQLite cursor class."""

    def __init__(self, executor, connection=None):
        """Create SQLite Cursor.

        :param executor: Executor
        :type executor: byte.executors.core.base.Executor

        :param connection: Connection
        :type connection: byte.executors.core.models.database.connection.Connection or None
        """
        super(PySqliteCursor, self).__init__(executor, connection=connection)

        # Create cursor
        self.instance = self.connection.instance.cursor()

    @property
    def columns(self):
        """Retrieve columns description from cursor."""
        return self.instance.description

    def execute(self, statement, parameters=None):
        """Execute statement."""
        log.debug('Execute: %r %r', statement, parameters)

        if not parameters:
            return self.instance.execute(statement)

        return self.instance.execute(statement, parameters)

    def close(self):
        """Close cursor."""
        super(PySqliteCursor, self).close()

        # Close cursor
        self.instance.close()
        self.instance = None

    def __iter__(self):
        return iter(self.instance)



from __future__ import absolute_import, division, print_function

from byte.collection import Collection
from byte.model import Model
from byte.property import Property
import byte.compilers.sqlite
import byte.executors.pysqlite

from hamcrest import *
from pysqlite2 import dbapi2 as sqlite3
import pytest


class User(Model):
    class Options:
        slots = True

    id = Property(int, primary_key=True)

    username = Property(str)
    password = Property(str)


def test_commit():
    """Test transaction is committed correctly."""
    users = Collection(User, 'pysqlite://:memory:?table=users', plugins=[
        byte.compilers.sqlite,
        byte.executors.pysqlite
    ])

    # Create table
    with users.executor.connection() as connection:
        with connection.cursor() as cursor:
            cursor.execute("""
                CREATE TABLE users (
                    id          INTEGER         PRIMARY KEY AUTOINCREMENT NOT NULL,
                    username    VARCHAR(255),
                    password    VARCHAR(255)
                );
            """)

    # Execute multiple queries inside transaction
    with users.transaction():
        users.insert().items({'id': 1, 'username': 'one', 'password': 'alpha'}).execute()
        users.insert().items({'id': 2, 'username': 'one', 'password': 'alpha'}).execute()
        users.insert().items({'id': 3, 'username': 'one', 'password': 'alpha'}).execute()

    # Ensure items were inserted
    assert_that(list(users.all()), has_length(3))

    # Validate executor state
    assert_that(users.executor.transactions, has_length(0))
    assert_that(users.executor.connections.active, equal_to(0))


def test_rollback():
    """Test transaction is committed correctly."""
    users = Collection(User, 'pysqlite://:memory:?table=users', plugins=[
        byte.compilers.sqlite,
        byte.executors.pysqlite
    ])

    # Create table
    with users.executor.connection() as connection:
        with connection.cursor() as cursor:
            cursor.execute("""
                CREATE TABLE users (
                    id          INTEGER         PRIMARY KEY AUTOINCREMENT NOT NULL,
                    username    VARCHAR(255),
                    password    VARCHAR(255)
                );
            """)

    # Insert multiple items (with conflict) inside transaction
    with pytest.raises(sqlite3.IntegrityError):
        with users.transaction():
            users.insert().items({'id': 1, 'username': 'one', 'password': 'alpha'}).execute()
            users.insert().items({'id': 2, 'username': 'one', 'password': 'alpha'}).execute()
            users.insert().items({'id': 1, 'username': 'one', 'password': 'alpha'}).execute()
            users.insert().items({'id': 3, 'username': 'one', 'password': 'alpha'}).execute()

    # Ensure items were inserted
    assert_that(list(users.all()), has_length(0))

    # Validate executor state
    assert_that(users.executor.transactions, has_length(0))
    assert_that(users.executor.connections.active, equal_to(0))


def test_already_inside_transaction():
    """Ensure an exception is raised while attempting to create a transaction while already in a transaction."""
    users = Collection(User, 'pysqlite://:memory:?table=users', plugins=[
        byte.compilers.sqlite,
        byte.executors.pysqlite
    ])

    # Insert multiple items (with conflict) inside transaction
    with pytest.raises(Exception):
        with users.transaction():
            with users.transaction():
                pass



"""byte-pysqlite - executor package."""
from __future__ import absolute_import, division, print_function

from byte.executors.pysqlite.main import PySqliteExecutor

__all__ = (
    'PySqliteExecutor',
)



from __future__ import absolute_import, division, print_function

from byte.collection import Collection
from byte.model import Model
from byte.property import Property
import byte.compilers.sqlite
import byte.executors.pysqlite

from hamcrest import *


class User(Model):
    class Options:
        slots = True

    id = Property(int, primary_key=True)

    username = Property(str)
    password = Property(str)


def test_single():
    """Test single item can be inserted into a database."""
    users = Collection(User, 'pysqlite://:memory:?table=users', plugins=[
        byte.compilers.sqlite,
        byte.executors.pysqlite
    ])

    # Create table
    with users.executor.connection() as connection:
        with connection.cursor() as cursor:
            cursor.execute("""
                CREATE TABLE users (
                    id          INTEGER         PRIMARY KEY AUTOINCREMENT NOT NULL,
                    username    VARCHAR(255),
                    password    VARCHAR(255)
                );
            """)

    # Insert item
    users.insert().items(
        {'username': 'one', 'password': 'alpha'}
    ).execute()

    # Validate items
    assert_that(list(users.all()), only_contains(
        has_properties({
            'username': 'one',
            'password': 'alpha'
        })
    ))


def test_multiple():
    """Test multiple items can be inserted into a database."""
    users = Collection(User, 'pysqlite://:memory:?table=users', plugins=[
        byte.compilers.sqlite,
        byte.executors.pysqlite
    ])

    # Create table
    with users.executor.connection() as connection:
        with connection.cursor() as cursor:
            cursor.execute("""
                CREATE TABLE users (
                    id          INTEGER         PRIMARY KEY AUTOINCREMENT NOT NULL,
                    username    VARCHAR(255),
                    password    VARCHAR(255)
                );
            """)

    # Insert items
    users.insert().items(
        {'username': 'one', 'password': 'alpha'},
        {'username': 'two', 'password': 'beta'},
        {'username': 'three', 'password': 'charlie'}
    ).execute()

    # Validate items
    assert_that(users.all(), only_contains(
        has_properties({
            'username': 'one',
            'password': 'alpha'
        }),
        has_properties({
            'username': 'two',
            'password': 'beta'
        }),
        has_properties({
            'username': 'three',
            'password': 'charlie'
        })
    ))



"""byte-pysqlite - executor tasks module."""
from __future__ import absolute_import, division, print_function

from byte.core.models import Task, ReadTask, SelectTask, WriteTask

import logging

log = logging.getLogger(__name__)


class PySqliteTask(Task):
    """PySQLite task base class."""

    def __init__(self, executor, statements):
        """Create SQLite executor task.

        :param executor: Executor
        :type executor: byte.executors.core.base.Executor

        :param statements: SQLite Statements
        :type statements: list of (str, tuple)
        """
        super(PySqliteTask, self).__init__(executor)

        self.statements = statements


class PySqliteReadTask(ReadTask, PySqliteTask):
    """PySQLite read task class."""

    def __init__(self, executor, statements):
        """Create SQLite read task.

        :param executor: Executor
        :type executor: byte.executors.core.base.Executor

        :param statements: SQLite Statements
        :type statements: list of (str, tuple)
        """
        super(PySqliteReadTask, self).__init__(executor, statements)

        self.cursor = self.executor.cursor()

    def execute(self):
        """Execute task."""
        for operation in self.statements:
            if not isinstance(operation, tuple) or len(operation) != 2:
                raise ValueError('Invalid statement returned from compiler: %s' % (operation,))

            self.cursor.execute(*operation)

        return self

    def close(self):
        """Close task."""
        self.cursor.close()


class PySqliteSelectTask(SelectTask, PySqliteReadTask):
    """PySQLite select task class."""

    def items(self):
        """Retrieve items from task."""
        # Parse items from cursor
        for row in self.cursor:
            yield self.model.from_plain(
                self._build_item(row),
                translate=True
            )

        # Close cursor
        self.close()

    def _build_item(self, row):
        data = {}

        for i, column in enumerate(self.cursor.columns):
            data[column[0]] = row[i]

        return data


class PySqliteWriteTask(WriteTask, PySqliteTask):
    """PySQLite write task class."""

    def __init__(self, executor, statements):
        """Create SQLite write task.

        :param executor: Executor
        :type executor: byte.executors.core.base.Executor

        :param statements: SQLite Statements
        :type statements: list of (str, tuple)
        """
        super(PySqliteWriteTask, self).__init__(executor, statements)

        # Retrieve transaction
        self.transaction = self.executor.transaction()

    def execute(self):
        """Execute task."""
        with self.transaction:
            for operation in self.statements:
                if not isinstance(operation, tuple) or len(operation) != 2:
                    raise ValueError('Invalid statement returned from compiler: %s' % (operation,))

                self.transaction.execute(*operation)

        return self

    def close(self):
        """Close task."""
        pass


class PySqliteInsertTask(PySqliteWriteTask):
    """PySQLite insert task class."""

    pass



from __future__ import absolute_import, division, print_function

import os
import subprocess


def detect_environment():
    if os.environ.get('TRAVIS_BUILD_ID'):
        return 'travis'

    return 'release'


def install():
    environment = os.environ.get('BYTE_ENVIRONMENT') or detect_environment()

    if not environment:
        raise Exception('Unable to detect environment, and no environment was provided')

    if environment == 'development':
        return install_development()

    if environment == 'release':
        return install_release()

    if environment == 'travis':
        return install_travis()

    raise Exception('Unknown environment: %s' % (environment,))


def install_development():
    # Install byte packages
    pip_upgrade(
        os.path.abspath('../byte'),
        os.path.abspath('../byte-sqlite')
    )

    # Install additional requirements
    pip_install(
        '-rrequirements.txt',
        '-rtests/requirements.txt'
    )


def install_release():
    # Install requirements
    pip_install(
        '-rrequirements.txt',
        '-rtests/requirements.txt'
    )


def install_travis():
    branch = os.environ.get('TRAVIS_BRANCH')

    if not branch:
        raise Exception('"TRAVIS_BRANCH" environment variable hasn\'t been defined')

    if branch != 'master':
        branch = 'develop'

    # Install byte packages
    pip_upgrade(
        'git+https://github.com/fuzeman/byte.git@%s' % (branch,),
        'git+https://github.com/fuzeman/byte-sqlite.git@%s' % (branch,)
    )

    # Install additional requirements
    pip_install(
        '-rrequirements.txt',
        '-rtests/requirements.txt'
    )


def pip_install(*args):
    execute('pip', 'install', *args)


def pip_upgrade(*args):
    execute('pip', 'install', '--upgrade', *args)


def execute(*args):
    p = subprocess.Popen(list(args), shell=False)
    p.communicate()


if __name__ == '__main__':
    install()
